import torch
import torch.nn as nn
from einops import rearrange, repeat

from ...common.base_model import BaseClassificationModel
from ...decoder.mlp import MLPDecoder
from ...encoder.embedding import LinearVideoEmbedding, PosEmbedding, TubeletEmbedding
from ...encoder.vanilla import VanillaEncoder
from ...encoder.vivit import ViViTEncoder
from ...utils.registry import MODEL_REGISTRY
from ...utils.utils import pair


@MODEL_REGISTRY.register()
class ViViTModel2(BaseClassificationModel):
    """
    Model 2 implementation of  A Video vision Transformer -
    Parameters
    ----------
    img_size:int
        Size of single frame/ image in video
    in_channels:int
        Number of channels
    patch_size: int
        Patch size
    embedding_dim: int
        Embedding dimension of a patch
    num_frames:int
        Number of seconds in each Video
    depth:int
        Number of encoder layers
    num_heads:int
        Number of attention heads
    head_dim:int
        Dimension of head
    num_classes:int
        Number of classes
    mlp_dim: int
        Dimension of hidden layer
    pool: str
        Pooling operation,must be one of {"cls","mean"},default is "cls"
    p_dropout:float
        Dropout probability
    attn_dropout:float
        Dropout probability
    drop_path_rate:float
        Stochastic drop path rate
    """

    def __init__(
        self,
        img_size,
        in_channels,
        patch_size,
        embedding_dim,
        num_frames,
        depth,
        num_heads,
        head_dim,
        num_classes,
        mlp_dim=None,
        pool="cls",
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path_rate=0.02,
    ):
        super(ViViTModel2, self).__init__(
            img_size=img_size,
            in_channels=in_channels,
            patch_size=patch_size,
            pool=pool,
        )

        patch_dim = in_channels * patch_size ** 2
        self.patch_embedding = LinearVideoEmbedding(
            embedding_dim=embedding_dim,
            patch_height=patch_size,
            patch_width=patch_size,
            patch_dim=patch_dim,
        )

        self.pos_embedding = PosEmbedding(
            shape=[num_frames, self.num_patches + 1], dim=embedding_dim, drop=p_dropout
        )

        self.space_token = nn.Parameter(
            torch.randn(1, 1, embedding_dim)
        )  # this is similar to using cls token in vanilla vision transformer
        self.spatial_transformer = VanillaEncoder(
            embedding_dim=embedding_dim,
            depth=depth,
            num_heads=num_heads,
            head_dim=head_dim,
            mlp_dim=mlp_dim,
            p_dropout=p_dropout,
            attn_dropout=attn_dropout,
            drop_path_rate=drop_path_rate,
        )

        self.time_token = nn.Parameter(torch.randn(1, 1, embedding_dim))
        self.temporal_transformer = VanillaEncoder(
            embedding_dim=embedding_dim,
            depth=depth,
            num_heads=num_heads,
            head_dim=head_dim,
            mlp_dim=mlp_dim,
            p_dropout=p_dropout,
            attn_dropout=attn_dropout,
            drop_path_rate=drop_path_rate,
        )

        self.decoder = MLPDecoder(
            config=[
                embedding_dim,
            ],
            n_classes=num_classes,
        )

    def forward(self, x):

        x = self.patch_embedding(x)

        (
            b,
            t,
            n,
            d,
        ) = x.shape  # shape of x will be number of videos,time,num_frames,embedding dim
        cls_space_tokens = repeat(self.space_token, "() n d -> b t n d", b=b, t=t)

        x = nn.Parameter(torch.cat((cls_space_tokens, x), dim=2))
        x = self.pos_embedding(x)

        x = rearrange(x, "b t n d -> (b t) n d")
        x = self.spatial_transformer(x)
        x = rearrange(x[:, 0], "(b t) ... -> b t ...", b=b)

        cls_temporal_tokens = repeat(self.time_token, "() n d -> b n d", b=b)
        x = torch.cat((cls_temporal_tokens, x), dim=1)

        x = self.temporal_transformer(x)

        x = x.mean(dim=1) if self.pool == "mean" else x[:, 0]

        x = self.decoder(x)

        return x


# model 3
@MODEL_REGISTRY.register()
class ViViTModel3(BaseClassificationModel):
    """
    model 3 of A video Vision Trnasformer- https://arxiv.org/abs/2103.15691

    Parameters
    ----------
    img_size:int or tuple[int]
        size of a frame
    patch_t:int
        Temporal length of single tube/patch in tubelet embedding
    patch_h:int
        Height  of single tube/patch in tubelet embedding
    patch_w:int
        Width  of single tube/patch in tubelet embedding
    in_channels: int
        Number of input channels, default is 3
    num_classes:int
        Number of classes
    num_frames :int
        Number of seconds in each Video
    embedding_dim:int
        Embedding dimension of a patch
    depth:int
        Number of Encoder layers
    num_heads: int
        Number of attention heads
    head_dim:int
        Dimension of attention head
    p_dropout:float
        Dropout rate/probability, default is 0.0
    mlp_dim: int
        Hidden dimension, optional
    """

    def __init__(
        self,
        img_size,
        patch_t,
        patch_h,
        patch_w,
        in_channels,
        num_classes,
        num_frames,
        embedding_dim,
        depth,
        num_heads,
        head_dim,
        p_dropout,
        mlp_dim=None,
    ):

        super(ViViTModel3, self).__init__(
            in_channels=in_channels,
            patch_size=(patch_h, patch_w),
            pool="mean",
            img_size=img_size,
        )
        h, w = pair(img_size)
        self.tubelet_embedding = TubeletEmbedding(
            embedding_dim=embedding_dim,
            tubelet_t=patch_t,
            tubelet_h=patch_h,
            tubelet_w=patch_w,
            in_channels=in_channels,
        )

        self.pos_embbedding = PosEmbedding(
            shape=[num_frames // patch_t, (h * w) // (patch_w * patch_h)],
            dim=embedding_dim,
        )
        self.encoder = ViViTEncoder(
            dim=embedding_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            p_dropout=p_dropout,
            depth=depth,
            hidden_dim=mlp_dim,
        )

        self.decoder = MLPDecoder(
            config=[
                embedding_dim,
            ],
            n_classes=num_classes,
        )

    def forward(self, x):

        x = self.tubelet_embedding(x)
        x = self.pos_embbedding(x)
        x = self.encoder(x)
        x = x.mean(dim=1)
        x = self.decoder(x)

        return x



from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange

from ..utils import ATTENTION_REGISTRY


@ATTENTION_REGISTRY.register()
class ConvVTAttention(nn.Module):
    """
    Attention with Convolutional Projection

    Parameters:
    ------------
    dim_in: int
        Dimension of input tensor
    dim_out: int
        Dimension of output tensor
    num_heads: int
        Number of heads in attention
    img_size: int
        Size of image
    attn_dropout: float
        Probability of dropout in attention
    proj_dropout: float
        Probability of dropout in convolution projection
    method: str ('dw_bn' for depth-wise convolution and batch norm, 'avg' for average pooling)
        Method of projection
    kernel_size: int
        Size of kernel
    stride_kv: int
        Size of stride for key value
    stride_q: int
        Size of stride for query
    padding_kv: int
        Padding for key value
    padding_q: int
        Padding for query
    with_cls_token: bool
        Whether to include classification token
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        num_heads,
        img_size,
        attn_dropout=0.0,
        proj_dropout=0.0,
        method="dw_bn",
        kernel_size=3,
        stride_kv=1,
        stride_q=1,
        padding_kv=1,
        padding_q=1,
        with_cls_token=False,
    ):
        super().__init__()
        self.stride_kv = stride_kv
        self.stride_q = stride_q
        self.with_cls_token = with_cls_token
        self.dim = dim_out
        self.num_heads = num_heads
        self.scale = dim_out ** -0.5
        self.h, self.w = img_size, img_size
        self.conv_proj_q = self._build_projection(
            dim_in, kernel_size, padding_q, stride_q, method
        )
        self.conv_proj_k = self._build_projection(
            dim_in, kernel_size, padding_kv, stride_kv, method
        )
        self.conv_proj_v = self._build_projection(
            dim_in, kernel_size, padding_kv, stride_kv, method
        )

        self.proj_q = nn.Linear(dim_in, dim_out, bias=False)
        self.proj_k = nn.Linear(dim_in, dim_out, bias=False)
        self.proj_v = nn.Linear(dim_in, dim_out, bias=False)

        self.attn_drop = nn.Dropout(attn_dropout)
        self.proj = nn.Linear(dim_out, dim_out)
        self.proj_drop = nn.Dropout(proj_dropout)

    def _build_projection(self, dim_in, kernel_size, padding, stride, method):
        if method == "dw_bn":
            proj = nn.Sequential(
                OrderedDict(
                    [
                        (
                            "conv",
                            nn.Conv2d(
                                dim_in,
                                dim_in,
                                kernel_size=kernel_size,
                                padding=padding,
                                stride=stride,
                                bias=False,
                                groups=dim_in,
                            ),
                        ),
                        ("bn", nn.BatchNorm2d(dim_in)),
                    ]
                )
            )
        elif method == "avg":
            proj = nn.AvgPool2d(
                kernel_size=kernel_size, padding=padding, stride=stride, ceil_mode=True
            )
        else:
            raise ValueError("Unknown method ({})".format(method))

        return proj

    def forward_conv(self, x):
        if self.with_cls_token:
            cls_token, x = torch.split(x, [1, self.h * self.w], 1)

        x = rearrange(x, "b (h w) c -> b c h w", h=self.h, w=self.w)

        q = self.conv_proj_q(x)
        q = rearrange(q, "b c h w -> b (h w) c")

        k = self.conv_proj_k(x)
        k = rearrange(k, "b c h w -> b (h w) c")

        v = self.conv_proj_v(x)
        v = rearrange(v, "b c h w -> b (h w) c")

        if self.with_cls_token:
            q = torch.cat((cls_token, q), dim=1)
            k = torch.cat((cls_token, k), dim=1)
            v = torch.cat((cls_token, v), dim=1)

        return q, k, v

    def forward(self, x):
        q, k, v = self.forward_conv(x)

        q = rearrange(self.proj_q(q), "b t (h d) -> b h t d", h=self.num_heads)
        k = rearrange(self.proj_k(k), "b t (h d) -> b h t d", h=self.num_heads)
        v = rearrange(self.proj_v(v), "b t (h d) -> b h t d", h=self.num_heads)

        attn_score = torch.einsum("bhlk,bhtk->bhlt", [q, k]) * self.scale
        attn = F.softmax(attn_score, dim=-1)
        attn = self.attn_drop(attn)

        x = torch.einsum("bhlt,bhtv->bhlv", [attn, v])
        x = rearrange(x, "b h t d -> b t (h d)")

        x = self.proj(x)
        x = self.proj_drop(x)

        return x



import torch
import torch.nn as nn
from torchvision.transforms.functional import resize

from ....utils import DECODER_REGISTRY


class DoubleConv(nn.Module):
    """
    Module consisting of two convolution layers and activations
    """

    def __init__(
        self,
        in_channels,
        out_channels,
    ):
        super(DoubleConv, self).__init__()

        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, 3, 1, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, 3, 1, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.conv(x)


@DECODER_REGISTRY.register()
class SegmentationHead(nn.Module):
    """
    U-net like up-sampling block
    """

    def __init__(
        self,
        out_channels=1,
        embed_dims=[64, 128, 256, 512],
    ):
        super(SegmentationHead, self).__init__()

        self.ups = nn.ModuleList()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        for feature in reversed(embed_dims):
            self.ups.append(
                nn.ConvTranspose2d(
                    feature * 2,
                    feature,
                    kernel_size=2,
                    stride=2,
                )
            )
            self.ups.append(DoubleConv(feature * 2, feature))

        self.bottleneck = DoubleConv(embed_dims[-1], embed_dims[-1] * 2)
        self.conv1 = nn.Conv2d(embed_dims[0], out_channels, kernel_size=1)
        self.conv2 = nn.ConvTranspose2d(
            out_channels, out_channels, kernel_size=4, stride=4
        )

    def forward(self, skip_connections):

        x = self.bottleneck(skip_connections[-1])
        skip_connections = skip_connections[::-1]

        for idx in range(0, len(self.ups), 2):

            x = self.ups[idx](x)
            skip_connection = skip_connections[idx // 2]

            if x.shape != skip_connection.shape:
                x = resize(x, size=skip_connection.shape[2:])

            concat_skip = torch.cat((skip_connection, x), dim=1)
            x = self.ups[idx + 1](concat_skip)

        x = self.conv1(x)

        return self.conv2(x)



import torch
import torch.nn as nn

from vformer.utils import MODEL_REGISTRY

models = MODEL_REGISTRY.get_list()
img_3channels_256 = torch.randn(2, 3, 256, 256)
img_3channels_224 = torch.randn(4, 3, 224, 224)
img_1channels_224 = torch.randn(2, 1, 224, 224)


def test_VanillaViT():

    model = MODEL_REGISTRY.get("VanillaViT")(
        img_size=256, patch_size=32, n_classes=10, in_channels=3
    )
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model

    model = MODEL_REGISTRY.get("VanillaViT")(
        img_size=256,
        patch_size=32,
        n_classes=10,
        embedding_dim=1024,
        decoder_config=(1024, 512),
    )
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model


def test_SwinTransformer():

    model = MODEL_REGISTRY.get("SwinTransformer")(
        img_size=224,
        patch_size=4,
        in_channels=3,
        n_classes=1000,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=7,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        p_dropout=0.0,
        attn_dropout=0.0,
        drop_path_rate=0.1,
        norm_layer=nn.LayerNorm,
        ape=False,
        patch_norm=True,
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)
    del model

    # tiny_patch4_window7_224
    model = MODEL_REGISTRY.get("SwinTransformer")(
        img_size=224,
        patch_size=4,
        in_channels=3,
        n_classes=10,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=7,
        p_dropout=0.2,
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    # tiny_c24_patch4_window8_256
    model = MODEL_REGISTRY.get("SwinTransformer")(
        img_size=256,
        patch_size=4,
        in_channels=3,
        n_classes=10,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[4, 8, 16, 32],
        window_size=8,
        p_dropout=0.2,
    )
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model

    # for greyscale image
    model = MODEL_REGISTRY.get("SwinTransformer")(
        img_size=224,
        patch_size=4,
        in_channels=1,
        n_classes=10,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=7,
        p_dropout=0.2,
    )
    out = model(img_1channels_224)
    assert out.shape == (2, 10)
    del model

    # testing for decoder_config parameter
    model = MODEL_REGISTRY.get("SwinTransformer")(
        img_size=224,
        patch_size=4,
        in_channels=3,
        n_classes=10,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=7,
        p_dropout=0.2,
        decoder_config=(768, 256, 10, 2),
    )
    out = model(img_3channels_224)
    del model
    assert out.shape == (4, 10)

    # ape=false
    model = MODEL_REGISTRY.get("SwinTransformer")(
        img_size=224,
        patch_size=4,
        in_channels=3,
        n_classes=10,
        embedding_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=7,
        p_dropout=0.2,
        decoder_config=(768, 256, 10, 2),
        ape=False,
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model


def test_CrossVit():

    model = MODEL_REGISTRY.get("CrossViT")(256, 16, 64, 10)
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model

    model = MODEL_REGISTRY.get("CrossViT")(
        256,
        16,
        64,
        10,
        decoder_config_s=(1024, 256, 10),
        decoder_config_l=(1024, 256, 10),
    )
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model


def test_pvt():
    # classification
    model = MODEL_REGISTRY.get("PVTClassification")(
        patch_size=[7, 3, 3, 3],
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratio=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=nn.LayerNorm,
        depths=[2, 2, 2, 2],
        sr_ratios=[8, 4, 2, 1],
        decoder_config=[512, 10],
        n_classes=10,
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    model = MODEL_REGISTRY.get("PVTClassification")(
        patch_size=[7, 3, 3, 3],
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratio=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=nn.LayerNorm,
        depths=[2, 2, 2, 2],
        sr_ratios=[8, 4, 2, 1],
        decoder_config=512,
        n_classes=10,
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    model = MODEL_REGISTRY.get("PVTClassificationV2")(linear=False)
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)
    del model

    model = MODEL_REGISTRY.get("PVTClassificationV2")(n_classes=10)
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    model = MODEL_REGISTRY.get("PVTClassificationV2")(n_classes=10)
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    model = MODEL_REGISTRY.get("PVTClassification")(n_classes=12)
    out = model(img_3channels_224)
    assert out.shape == (4, 12)
    del model

    model = MODEL_REGISTRY.get("PVTClassificationV2")(
        embedding_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratio=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=nn.LayerNorm,
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        linear=True,
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)

    # segmentation
    model = MODEL_REGISTRY.get("PVTSegmentation")()
    outs = model(img_3channels_224)
    assert outs.shape == (
        4,
        1,
        224,
        224,
    ), f"expected: {(4,1,224,224)}, got : {outs.shape}"
    del model

    model = MODEL_REGISTRY.get("PVTSegmentation")()
    outs = model(img_3channels_256)
    assert outs.shape == (
        2,
        1,
        256,
        256,
    ), f"expected: {(4,1,256,256)}, got : {outs.shape}"
    del model

    model = MODEL_REGISTRY.get("PVTSegmentation")()
    outs = model(img_3channels_256)
    assert outs.shape == (
        2,
        1,
        256,
        256,
    ), f"expected: {(4,1,256,256)}, got : {outs.shape}"
    del model

    model = MODEL_REGISTRY.get("PVTSegmentationV2")(return_pyramid=False)
    outs = model(img_3channels_224)
    assert outs.shape == (
        4,
        1,
        224,
        224,
    ), f"expected: {(4,1,224,224)}, got : {outs.shape}"
    del model

    model = MODEL_REGISTRY.get("PVTSegmentationV2")(return_pyramid=True)
    out = model(img_3channels_224)

    model = MODEL_REGISTRY.get("PVTSegmentationV2")(return_pyramid=False)
    outs = model(img_3channels_256)
    assert outs.shape == (
        2,
        1,
        256,
        256,
    ), f"expected: {(4,1,256,256)}, got : {outs.shape}"
    del model

    # detection
    model = MODEL_REGISTRY.get("PVTDetection")()
    outs = model(img_3channels_224)
    del model

    model = MODEL_REGISTRY.get("PVTDetectionV2")()
    outs = model(img_3channels_224)
    del model


def test_cvt():

    model = MODEL_REGISTRY.get("CVT")(img_size=256, patch_size=4, in_channels=3)
    out = model(img_3channels_256)
    assert out.shape == (2, 1000)
    del model

    model = MODEL_REGISTRY.get("CVT")(
        img_size=224,
        patch_size=4,
        in_channels=3,
        seq_pool=False,
        embedding_dim=768,
        num_heads=1,
        mlp_ratio=4.0,
        n_classes=10,
        p_dropout=0.5,
        attn_dropout=0.3,
        drop_path=0.2,
        positional_embedding="sine",
        decoder_config=(768, 12024, 512, 256, 128, 64, 32),
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    model = MODEL_REGISTRY.get("CVT")(
        img_size=224,
        in_channels=3,
        patch_size=4,
        positional_embedding="none",
        seq_pool=False,
        decoder_config=None,
    )
    f = model(img_3channels_224)
    assert f.shape == (4, 1000)
    del model

    model = MODEL_REGISTRY.get("CVT")(
        img_size=224,
        in_channels=3,
        patch_size=4,
        positional_embedding="none",
        seq_pool=True,
        decoder_config=768,
    )
    f = model(img_3channels_224)
    assert f.shape == (4, 1000)
    del model


def test_cct():

    model = MODEL_REGISTRY.get("CCT")(img_size=256, patch_size=4, in_channels=3)
    out = model(img_3channels_256)
    assert out.shape == (2, 1000)
    del model

    model = MODEL_REGISTRY.get("CCT")(
        img_size=224,
        patch_size=4,
        in_channels=3,
        seq_pool=False,
        embedding_dim=768,
        num_heads=1,
        mlp_ratio=4.0,
        n_classes=10,
        p_dropout=0.5,
        attn_dropout=0.3,
        drop_path=0.2,
        positional_embedding="sine",
        decoder_config=(768, 12024, 512, 256, 128, 64, 32),
    )
    out = model(img_3channels_224)
    assert out.shape == (4, 10)
    del model

    model = MODEL_REGISTRY.get("CCT")(
        img_size=224,
        in_channels=3,
        patch_size=4,
        positional_embedding="none",
        seq_pool=False,
        decoder_config=None,
    )
    f = model(img_3channels_224)
    assert f.shape == (4, 1000)
    del model

    model = MODEL_REGISTRY.get("CCT")(
        img_size=224,
        in_channels=3,
        patch_size=4,
        positional_embedding="none",
        seq_pool=True,
        decoder_config=768,
    )
    f = model(img_3channels_224)
    assert f.shape == (4, 1000)
    del model


def test_visformer():

    model = MODEL_REGISTRY.get("Visformer_S")(224, 1000)
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)
    del model

    model = MODEL_REGISTRY.get("Visformer_Ti")(224, 1000)
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)
    del model

    model = MODEL_REGISTRY.get("VisformerV2_S")(224, 1000)
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)
    del model

    model = MODEL_REGISTRY.get("VisformerV2_Ti")(224, 1000)
    out = model(img_3channels_224)
    assert out.shape == (4, 1000)
    del model


def test_dpt():
    img = torch.randn(4, 3, 384, 384)
    model = MODEL_REGISTRY.get("DPTDepth")(
        "vitb16",
        enable_attention_hooks=True,
    )
    del model
    model = MODEL_REGISTRY.get("DPTDepth")("vitl16")
    del model

    model = MODEL_REGISTRY.get("DPTDepth")("vitl16", invert=True, readout="ignore")
    del model

    model = MODEL_REGISTRY.get("DPTDepth")("vitb16", invert=True, readout="add")
    del model
    """
    only initialisation of large models; no forward pass here because these models are
    very large and github CI pipeline wont be able to handle them.
    forward pass will be done on vit tiny model.
    """

    model = MODEL_REGISTRY.get("DPTDepth")(
        "vit_tiny", enable_attention_hooks=True, channels_last=True
    )
    out = model(img)
    assert out.shape == (4, 384, 384)
    del model

    model = MODEL_REGISTRY.get("DPTDepth")("vit_tiny", invert=True, readout="ignore")
    out = model(img)
    assert out.shape == (4, 384, 384)
    del model

    model = MODEL_REGISTRY.get("DPTDepth")("vit_tiny", readout="add", use_bn=True)
    out = model(img)
    assert out.shape == (4, 384, 384)
    del model

    model = MODEL_REGISTRY.get("DPTDepth")(
        "vit_tiny",
        readout="add",
        use_bn=True,
    )
    out = model(img)
    assert out.shape == (4, 384, 384)
    del model


def test_ConViT():

    model = MODEL_REGISTRY.get("ConViT")(
        img_size=256, patch_size=32, n_classes=10, in_channels=3
    )
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model

    model = MODEL_REGISTRY.get("ConViT")(
        img_size=256,
        patch_size=32,
        n_classes=10,
        embedding_dim=1024,
        decoder_config=(1024, 512),
    )
    out = model(img_3channels_256)
    assert out.shape == (2, 10)
    del model


def test_ConvVT():
    img = torch.randn(4, 3, 224, 224)
    model = MODEL_REGISTRY.get("ConvVT")()
    out = model(img)
    assert out.shape == torch.Size([4, 1000])
    del model

    model = MODEL_REGISTRY.get("ConvVT")(img_size=384)
    img2 = torch.randn(4, 3, 384, 384)
    out = model(img2)
    assert out.shape == torch.Size([4, 1000])
    del model


def test_ViViT():
    test_tensor1 = torch.randn([1, 16, 3, 224, 224])
    test_tensor2 = torch.randn([3, 16, 3, 224, 224])

    model = MODEL_REGISTRY.get("ViViTModel2")(
        img_size=224,
        in_channels=3,
        patch_size=16,
        embedding_dim=192,
        depth=4,
        num_heads=3,
        head_dim=64,
        num_frames=1,
        num_classes=10,
    )

    out = model(test_tensor1)
    assert out.shape == (1, 10)

    out = model(test_tensor2)
    assert out.shape == (3, 10)
    del model

    model = MODEL_REGISTRY.get("ViViTModel3")(
        num_frames=32,
        img_size=(64, 64),
        patch_t=8,
        patch_h=4,
        patch_w=4,
        num_classes=10,
        embedding_dim=512,
        depth=3,
        num_heads=4,
        head_dim=32,
        p_dropout=0.0,
        in_channels=3,
    )
    test_tensor3 = torch.randn(32, 32, 3, 64, 64)
    logits = model(test_tensor3)
    assert logits.shape == (32, 10)
    del model

    model = MODEL_REGISTRY.get("ViViTModel3")(
        num_frames=16,
        img_size=(64, 64),
        patch_t=8,
        patch_h=4,
        patch_w=4,
        num_classes=10,
        embedding_dim=512,
        depth=3,
        num_heads=4,
        head_dim=32,
        p_dropout=0.0,
        in_channels=1,
    )

    test_tensor4 = torch.randn(7, 16, 1, 64, 64)
    logits = model(test_tensor4)
    assert logits.shape == (7, 10)



# Video Patch Embedding
# data Dim is in following format - Batch,Time,Channels,Height,Width

# As discussed in paper; this embedding is just Like Vanilla Embedding where we take non overlapping image patches and map them into multidimension embedding

import torch
import torch.nn as nn
from einops.layers.torch import Rearrange


class LinearVideoEmbedding(nn.Module):
    """

    Parameters
    ----------
    embedding_dim: int
        Dimension of the resultant embedding
    patch_height: int
        Height of the patch
    patch_width: int
        Width of the patch

    """

    def __init__(
        self,
        embedding_dim,
        patch_height,
        patch_width,
        patch_dim,
    ):

        super().__init__()
        self.patch_embedding = nn.Sequential(
            Rearrange(
                "b t c (h ph) (w pw) -> b t (h w) (ph pw c)",
                ph=patch_height,
                pw=patch_width,
            ),
            nn.Linear(patch_dim, embedding_dim),
        )

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            Input tensor

        Returns
        ----------
        torch.Tensor
            Returns patch embeddings of size `embedding_dim`

        """

        return self.patch_embedding(x)


#
class TubeletEmbedding(nn.Module):
    """

    Parameters
    ----------
    embedding_dim: int
        Dimension of the resultant embedding
    tubelet_t: int
        Temporal length of single tube/patch
    tubelet_h: int
        Heigth  of single tube/patch
    tubelet_w: int
        Width of single tube/patch

    """

    def __init__(self, embedding_dim, tubelet_t, tubelet_h, tubelet_w, in_channels):
        super(TubeletEmbedding, self).__init__()
        tubelet_dim = in_channels * tubelet_h * tubelet_w * tubelet_t
        self.tubelet_embedding = nn.Sequential(
            Rearrange(
                "b  (t pt) c (h ph) (w pw) -> b t (h w) (pt ph pw c)",
                pt=tubelet_t,
                ph=tubelet_h,
                pw=tubelet_w,
            ),
            nn.Linear(tubelet_dim, embedding_dim),
        )

    def forward(self, x):
        """

        Parameters
        ----------
        x: Torch.tensor
            Input tensor

        """
        return self.tubelet_embedding(x)
